#include "ObjectManager.h"

#include <KDebug.h>

#include "Collisions.h"
#include "Debug.h"
#include "generated\DebugSettings.h"
#include "FunctionStubs.h"
#include "HUDManager.h"
#include "LevelManager.h"
#include "ScrollManager.h"
#include "MathUtils.h"
#include "exported\ExportedGGAnimations.h"
#include "PlayerManager.h"
#include "vdp_spr.h"
#include "VDP_Extra.h"

extern u16 spriteNum;

#define MAX_OBJECTS 60

ObjectStruct    objects[MAX_OBJECTS];
u16             numActiveObjects = 0;

ObjectStruct* firstVisibleObject = NULL;
ObjectStruct* firstObjectOfType[MAX_OBJECT_TYPES];
ObjectStruct* lastObjectOfType[MAX_OBJECT_TYPES];

u16 (*ActiveAreaIsWithinSpan)(ObjectStruct* object);

u16 ActiveAreaWithinHorizontalScreenSpan(ObjectStruct* object);
u16 ActiveAreaWithinVerticalScreenSpan(ObjectStruct* object);

void ObjectManager_Init()
{
    numActiveObjects = 0;

    PlayerManager_InitPosition(currentForegroundMap->playerStartX, currentForegroundMap->playerStartY);

    if (currentForegroundMap->mapType == MAPTYPE_HORIZONTAL ||
        currentForegroundMap->mapType == MAPTYPE_VDP)
    {
        ActiveAreaIsWithinSpan = ActiveAreaWithinHorizontalScreenSpan;
    }
    else if (currentForegroundMap->mapType == MAPTYPE_VERTICAL)
    {
        ActiveAreaIsWithinSpan = ActiveAreaWithinVerticalScreenSpan;
    }
}

void ObjectManager_ResetEnemyProjectileCollection()
{
    // go through visible projectile list and set everything to dead.
}

void SetupObjectList()
{

}

ObjectStruct* ObjectManager_CreateObject(u16                            objectType, 
                                         ObjectFunctionType             updateFunction, 
                                         ObjectFunctionType             destroyFunction,
                                         ObjectFunctionType             drawFunction,
                                         ObjectCollisionFunctionType    collisionFunction)
{
    if (numActiveObjects >= MAX_OBJECTS)
    {
        return NULL;
    }

    ObjectStruct* object = objects + numActiveObjects;

    object->updateFunction      = updateFunction;
    object->destroyFunction     = destroyFunction;
    object->drawFunction        = drawFunction;
    object->collisionFunction   = collisionFunction;
    object->objectState         = OBJECTSTATE_ACTIVE;
    object->objectType          = objectType;
    object->objectStreamerId    = -1;

    numActiveObjects++;

    return object;
} 

void DrawObjects(u16 objectType)
{
    if (*(firstObjectOfType + objectType)== NULL) return;

    ObjectStruct* object = *(firstObjectOfType + objectType);

    while (object != NULL)
    {
        object->drawFunction(object);
        object = (ObjectStruct*)object->nextVisibleObjectOfSameType;
    }

}

void ObjectManager_DrawObjects()
{
    MyVDP_resetSprites();

    HUDManager_Draw();

    DrawObjects(TYPE_PICKUP);
    DrawObjects(TYPE_EFFECT);
    DrawObjects(TYPE_PLAYER_PROJECTILE);
    DrawObjects(TYPE_ENEMY_PROJECTILE);

    playerObject->drawFunction(playerObject);

    DrawObjects(TYPE_PLATFORM);
    DrawObjects(TYPE_ENEMY);

    MyVDP_setLastSprite();
}

void ObjectManager_UpdateVDP()
{
    VDP_updateSprites();
}

void ProcessCollisions(u16 firstType, u16 secondType)
{
    if (*(firstObjectOfType + firstType )== NULL) return;
    if (*(firstObjectOfType + secondType) == NULL) return;

    ObjectStruct* objectOfFirstType = *(firstObjectOfType + firstType);
    ObjectStruct* objectOfSecondType = NULL;

    while (objectOfFirstType != NULL)
    {
        if (objectOfFirstType->healthPoints > 0 &&
            objectOfFirstType->objectState == OBJECTSTATE_ACTIVE)
        {
            objectOfSecondType = *(firstObjectOfType + secondType);

            while (objectOfSecondType != NULL)
            {
                if (objectOfSecondType->healthPoints > 0 &&
                    objectOfSecondType->objectState == OBJECTSTATE_ACTIVE)
                {
                    if (!(objectOfFirstType->x + objectOfFirstType->rectLeft   > objectOfSecondType->x + objectOfSecondType->rectRight ||
                          objectOfFirstType->x + objectOfFirstType->rectRight  < objectOfSecondType->x + objectOfSecondType->rectLeft ||
                          objectOfFirstType->y + objectOfFirstType->rectTop    > objectOfSecondType->y + objectOfSecondType->rectBottom ||
                          objectOfFirstType->y + objectOfFirstType->rectBottom < objectOfSecondType->y + objectOfSecondType->rectTop))
                    {
                        objectOfFirstType->collisionFunction(objectOfFirstType, objectOfSecondType);
                        objectOfSecondType->collisionFunction(objectOfSecondType, objectOfFirstType);
                    }
                }

                objectOfSecondType = objectOfSecondType->nextVisibleObjectOfSameType;
            }
        }

        objectOfFirstType = objectOfFirstType->nextVisibleObjectOfSameType;
    }
}

void ProcessCollisionsAgainstPlayer(u16 objectType)
{
    if (*(firstObjectOfType + objectType)== NULL) return;

    // no need to do anything if we're dead.
    if (playerObject->objectState != OBJECTSTATE_ACTIVE)
    {
        return;
    }

    ObjectStruct* firstObject = *(firstObjectOfType + objectType);

    while (firstObject != NULL)
    {
        if (firstObject->healthPoints > 0 &&
            firstObject->objectState == OBJECTSTATE_ACTIVE)
        {
            if (!(firstObject->x + firstObject->rectLeft   > playerObject->x + playerObject->rectRight ||
                  firstObject->x + firstObject->rectRight  < playerObject->x + playerObject->rectLeft ||
                  firstObject->y + firstObject->rectTop    > playerObject->y + playerObject->rectBottom ||
                  firstObject->y + firstObject->rectBottom < playerObject->y + playerObject->rectTop))
            {
                firstObject->collisionFunction(firstObject, playerObject);
                playerObject->collisionFunction(playerObject, firstObject);
            }

        }

        firstObject = firstObject->nextVisibleObjectOfSameType;
    }
}


void ProcessAllCollisions()
{
    ProcessCollisions(TYPE_PLAYER_PROJECTILE, TYPE_ENEMY);

    if (debugSettings.collisions == 1)
    {
        ProcessCollisionsAgainstPlayer(TYPE_ENEMY);
        ProcessCollisionsAgainstPlayer(TYPE_ENEMY_PROJECTILE);
        ProcessCollisionsAgainstPlayer(TYPE_PICKUP);
        ProcessCollisionsAgainstPlayer(TYPE_PLATFORM);
    }
}


u16 ActiveAreaWithinHorizontalScreenSpan(ObjectStruct* object)
{
    return !((object->activeAreaRectLeft   > foregroundScrollData.scrollx_vdp + SCREEN_WIDTH) ||
             (object->activeAreaRectRight  < foregroundScrollData.scrollx_vdp));
}

u16 ActiveAreaWithinVerticalScreenSpan(ObjectStruct* object)
{
    return !((object->activeAreaRectTop    > foregroundScrollData.scrolly_vdp + SCREEN_HEIGHT) ||
             (object->activeAreaRectBottom < foregroundScrollData.scrolly_vdp));
}


void UpdateObjects()
{
    ObjectStruct* tempObject = objects;
    ObjectStruct* lastObject = (objects + numActiveObjects);

    firstVisibleObject = NULL;

    memset(firstObjectOfType, NULL, sizeof(firstObjectOfType));
    memset(lastObjectOfType, NULL, sizeof(lastObjectOfType));

    ObjectStruct** firstObjectOfTypePointer = firstObjectOfType;
    ObjectStruct** lastObjectOfTypePointer = lastObjectOfType;

    u16 activeAreaIsWithinScreenSpan = FALSE;

    while (tempObject != lastObject)
    {
        tempObject->updateFunction(tempObject);

        // ObjectIsOnScreen is computed from the sprite position and not the
        // object position. So make sure to compute the sprite position during
        // the object's update.

        tempObject->visible = ObjectIsOnScreen(tempObject);
        activeAreaIsWithinScreenSpan = ActiveAreaIsWithinSpan(tempObject);

        // if both the object and its active area are off screen, then
        // prepare the object to be destroyed.
        if (!tempObject->visible && !activeAreaIsWithinScreenSpan)
        {
            //Print("Force delete");
            tempObject->objectState = OBJECTSTATE_INACTIVE;

            // if the object had been created by the object streamer, 
            // tell it it's being destroyed.
            if (tempObject->objectStreamerId >= 0)
            {
                ObjectStreamer_DestroyObject(tempObject);
            }
        }

        if (tempObject->objectState == OBJECTSTATE_INACTIVE)
        {
            tempObject->destroyFunction(tempObject); // what happens if they create any new objects here???

            numActiveObjects--; 

            // set the new last object
            lastObject = (objects + numActiveObjects);

            // swap the top of the stack with the current dead object.
            // this will eliminate any holes in the stack.
            if (tempObject != lastObject)
            {
                memcpy(tempObject, lastObject, sizeof(ObjectStruct));
            }

            // we don't move the tempObject pointer because we've copied a new object
            // to where the dead one is. So we want to reprocess it.
        }
        else if (tempObject->objectState == OBJECTSTATE_ACTIVE)
        {
            tempObject->lifetime++; // loops after 65535. Maximum of 18.2 minutes.

            if (tempObject->visible)
            {
                // update ordering of same type.
                u16 objectType = tempObject->objectType;

                if (*(firstObjectOfTypePointer + objectType) == NULL)
                {
                    *(firstObjectOfTypePointer + objectType) = tempObject;
                }

                if (*(lastObjectOfTypePointer + objectType) != NULL)
                {
                    (*(lastObjectOfTypePointer + objectType))->nextVisibleObjectOfSameType = tempObject;
                }

                tempObject->nextVisibleObjectOfSameType = NULL;
                *(lastObjectOfTypePointer + objectType) = tempObject;
            }

            tempObject++;
        }
    }
}

void ObjectManager_UpdateObjects()
{
    playerObject->updateFunction(playerObject);
    playerObject->lifetime++;

    UpdateObjects();

    ProcessAllCollisions();
}


void ComputeStandardActiveArea(ObjectStruct* object)
{
    // I could change this so that the active area aligns with spawn sections someday.

    #define STANDARD_PADDING 32

    object->activeAreaRectLeft = VIRTUAL_TO_PIXEL(object->x) - STANDARD_PADDING;
    object->activeAreaRectRight = VIRTUAL_TO_PIXEL(object->x) + object->animation->width + STANDARD_PADDING;
    object->activeAreaRectTop = VIRTUAL_TO_PIXEL(object->y) - STANDARD_PADDING;
    object->activeAreaRectBottom = VIRTUAL_TO_PIXEL(object->y) + object->animation->height + STANDARD_PADDING;
}

