#include "PlayerManager.h"

#include <genesis.h>
#include <kdebug.h>
#include <maths.h>
#include <vdp_spr.h>

#include "generated\AllObjects.h"

#include "GGAnimationDraw.h"
#include "Collisions.h"
#include "Config.h"
#include "Debug.h"
#include "generated\DebugSettings.h"
#include "EnemyTypes.h"
#include "exported\ExportedGenResSprites.h"
#include "exported\ExportedGGAnimations.h"
#include "GameOverScreen.h"
#include "HUDManager.h"
#include "JoypadManager.h"
#include "LevelManager.h"
#include "MathUtils.h"
#include "ObjectManager.h"
#include "objects\ObjectMovement.h"
#include "ScrollManager.h"
#include "StateManager.h"
#include "TileHeightMaps.h"
#include "VDP_Extra.h"

// game constants.
#define MAX_HEALTH_POINTS               4
#define JUMP_SPEED                      25
#define KNOCKBACK_SPEED                 20
#define KNOCKBACK_UP_SPEED              -16
#define DEATH_JUMP_SPEED                17
#define MOVEMENT_SPEED                  4 
#define DAMAGE_TIME                     10
#define INVINSIBILITY_TIME              45

// helper macros
#define SET_ANIMATION(animation) \
    body.currentVDPInfo = &Girl##animation##VDPInfo;\
    body.currentAnimation = &Girl##animation; \
    body.animationFrame = 0;\
    body.animationSpeedCounter = 0;


ObjectStruct player;
ObjectStruct* playerObject;
VDPInfo playerVDPInfo;
PlayerDataStruct playerData;

typedef void (*VoidFunctionType)();

u16 isEndingScreen;

VoidFunctionType updateState = NULL;
VoidFunctionType updatePhysics = NULL;
VoidFunctionType onPhysicsCollision = NULL;

// Girl body
struct
{
    const VDPInfo* currentVDPInfo;
    const GGAnimation* currentAnimation;
    u16 animationFrame;
    u16 animationSpeed;
    u16 animationSpeedCounter;
} body;

// various state tracking variables
u16 deathWaitTime = 0;
s16 jumpCounter = 0;
u16 deathFallCounter = 0;

void StartIdle();
void StartWalk();
void StartJump();
void StartFall();
void StartHappy();

void UpdateIdle();
void UpdateWalk();
void UpdateJump();
void UpdateFall();
void UpdateHappy();

void UpdateIdlePhysics();
void UpdateWalkPhysics();
void UpdateJumpPhysics();
void UpdateFallPhysics();

void StartIdleAnimation();
void StartWalkAnimation();
void StartJumpAnimation();
void StartHappyAnimation();

void PlayerManager_AfterDeath();


void PlayerManager_Jump()
{
    StartJump();
}

void PlayerManager_Happy()
{
    StartHappy();
}

void PlayerManager_CallOnPhysicsCollisionFunction()
{
    if (onPhysicsCollision != NULL)
    {
        onPhysicsCollision();
    }
}

void StartIdle()
{
    playerData.playerState = PLAYERSTATE_IDLE;

    StartIdleAnimation();

    // updates
    updateState = UpdateIdle;
    updatePhysics = UpdateIdlePhysics;
    onPhysicsCollision = NULL;

    // misc
    playerObject->speedx = 0;
}

void StartWalk()
{
    playerData.playerState = PLAYERSTATE_WALK;

    StartWalkAnimation();

    // updates
    updateState = UpdateWalk;
    updatePhysics = UpdateWalkPhysics;
    onPhysicsCollision = NULL;
}

void StartJump()
{
    playerData.playerState = PLAYERSTATE_JUMP;

    StartJumpAnimation();

    // updates
    updateState = UpdateJump;
    updatePhysics = UpdateJumpPhysics;
    onPhysicsCollision = StartFall;

    // misc
    jumpCounter = 0;
    playerObject->speedy = -JUMP_SPEED;
    playerData.platform = NULL;
}

void OnFallPhysicsCollision()
{
    StartIdle();
}

void StartFall()
{
    playerData.playerState = PLAYERSTATE_FALL;

    StartJumpAnimation();

    // updates
    updateState = UpdateFall;
    updatePhysics = UpdateFallPhysics;
    onPhysicsCollision = OnFallPhysicsCollision;
    
    // misc
    jumpCounter = 0;
    playerObject->speedy = 0;
}

void StartHappy()
{
    playerData.playerState = PLAYERSTATE_HAPPY;

    StartHappyAnimation();

    // updates
    updateState = UpdateHappy;
    updatePhysics = UpdateIdlePhysics;
    onPhysicsCollision = NULL;
    
    // misc
    jumpCounter = 0;
    playerObject->speedy = 0;
}

void UpdateIdle()
{
    if ((joystate & BUTTON_LEFT) ||
        (joystate & BUTTON_RIGHT))
    {
        StartWalk();
    }

    if (joystickState.aPressed ||
        joystickState.bPressed ||
        joystickState.cPressed)
    {
        StartJump();
        return;
    }
}

void UpdateLeftRight()
{
    playerObject->speedx = 0;

    s16 movementSpeed = MOVEMENT_SPEED;

    if (joystate & BUTTON_LEFT)
    {
        playerObject->speedx = -movementSpeed;
    }
    else if (joystate & BUTTON_RIGHT)
    {
        playerObject->speedx = movementSpeed;
    }

    if ((playerObject->speedx != 0))
    {
        playerObject->direction = (int)(playerObject->speedx < 0);
    }
}



void UpdateWalk()
{
    UpdateLeftRight();

    if (joystickState.aPressed ||
        joystickState.bPressed ||
        joystickState.cPressed)
    {
        StartJump();
        return;
    }

    if (playerObject->speedx == 0)
    {
        StartIdle();
    }
}

void UpdateJump()
{
    UpdateLeftRight();

    playerObject->speedy += 2;
    jumpCounter++;

    if (playerObject->speedy > 0)
    {
        StartFall();
    }
}

void UpdateFall()
{
    UpdateLeftRight();

    playerObject->speedy++;

    if (playerObject->speedy > JUMP_SPEED)
    {
        playerObject->speedy = JUMP_SPEED;
    }
}

void UpdateHappy()
{

}

void StartIdleAnimation()
{
    SET_ANIMATION(Idle);
}

void StartWalkAnimation()
{
    SET_ANIMATION(Walk);
}

void StartJumpAnimation()
{
    SET_ANIMATION(Jump);
}

void StartHappyAnimation()
{
    SET_ANIMATION(Happy);
}

const s16 touchPointLeftOffset = -PIXEL_TO_VIRTUAL(3);
const s16 touchPointRightOffset = PIXEL_TO_VIRTUAL(3);

void UpdatePlatformCheck()
{
    // if two sensors isn't in the platform's rect, then fall.
    if ((playerObject->x + touchPointRightOffset < playerData.platform->x + playerData.platform->rectLeft) ||
        (playerObject->x + touchPointLeftOffset > playerData.platform->x + playerData.platform->rectRight))
    {
        StartFall();
        playerData.platform = NULL;
    }
}

void UpdateGroundCheck()
{
    if ((GetTileType_Virtual(playerObject->x + touchPointRightOffset, playerObject->y + PIXEL_TO_VIRTUAL(PLAYER_PIXEL_HEIGHT)) == TILE_NONE) &&
        (GetTileType_Virtual(playerObject->x + touchPointLeftOffset, playerObject->y + PIXEL_TO_VIRTUAL(PLAYER_PIXEL_HEIGHT)) == TILE_NONE))
    {
        StartFall();
    }
}

void UpdateIdlePhysics()
{
    if (playerData.platform != NULL)
    {
        playerObject->speedx = playerData.platform->speedx;
        playerObject->speedy = playerData.platform->speedy;

        playerObject->x += playerObject->speedx;
        playerObject->y += playerObject->speedy;
    }
    else
    {
        UpdateGroundCheck();
    }
}

s16 WalkLeftRightCollisionCheck(ObjectStruct* objectStruct, s16 speedx)
{
    s16 playerY = VIRTUAL_TO_PIXEL(playerObject->y);
    s16 endXPosition = VIRTUAL_TO_PIXEL(playerObject->x + speedx);

    if (speedx > 0)
    {
        endXPosition += 4;

        if ((endXPosition > TILE_TO_PIXEL(currentForegroundMap->width - 1)) ||
            (IsTileSolid_Pixel(endXPosition, playerY) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 16) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 24) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 32) != 0))
        {
            endXPosition >>= 3;
            endXPosition <<= 3;
            playerObject->x = PIXEL_TO_VIRTUAL(endXPosition - 4);
            speedx = 0;
        }

    }
    else if (speedx < 0)
    {
        endXPosition -= 4;

        if ((endXPosition < 8) ||
            (IsTileSolid_Pixel(endXPosition, playerY) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 16) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 24) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 32) != 0))
        {
            endXPosition >>= 3;
            endXPosition <<= 3;
            endXPosition += 8;
            playerObject->x = PIXEL_TO_VIRTUAL(endXPosition + 4);
            speedx = 0;
        }

    }

    return speedx;
}

void UpdateLeftRightWalkPhysics()
{
    // if we're on a platform, we need to add its speed to ours.
    if (playerData.platform != NULL)
    {
        playerObject->speedx += playerData.platform->speedx;
        playerObject->y += playerData.platform->speedy;
    }

    UpdateObjectLeftRightWalkPhysics(playerObject, WalkLeftRightCollisionCheck, PLAYER_PIXEL_HEIGHT);
}

void JumpFallLeftRightCollisionCheck()
{
    s16 playerY = VIRTUAL_TO_PIXEL(playerObject->y);
    s16 endXPosition = VIRTUAL_TO_PIXEL(playerObject->x + playerObject->speedx);

    if (playerObject->speedx > 0)
    {
        endXPosition += 4;

        if ((endXPosition > TILE_TO_PIXEL(currentForegroundMap->width - 1)) ||
            (IsTileSolid_Pixel(endXPosition, playerY) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 16) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 24) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 32) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 39) != 0))
        {
            endXPosition >>= 3;
            endXPosition <<= 3;
            playerObject->x = PIXEL_TO_VIRTUAL(endXPosition - 4);
            playerObject->speedx = 0;
        }
    }
    else if (playerObject->speedx < 0)
    {
        endXPosition -= 4;

        if ((endXPosition < 8) ||
            (IsTileSolid_Pixel(endXPosition, playerY) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 16) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 24) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 32) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 39) != 0))
        {
            endXPosition >>= 3;
            endXPosition <<= 3;
            endXPosition += 8;
            playerObject->x = PIXEL_TO_VIRTUAL(endXPosition + 4);
            playerObject->speedx = 0;
        }

    }

    playerObject->x += playerObject->speedx;
}

void UpdateWalkPhysics()
{
    UpdateLeftRightWalkPhysics();

    if (playerData.platform == NULL)
    {
        UpdateGroundCheck();
    }
    else
    {
        UpdatePlatformCheck();
    }
}

void UpdateJumpPhysics()
{
    s16 startPosition = playerObject->y;

    if (startPosition >= 0)
    {
        u16 endPosition = (u16)(startPosition + playerObject->speedy);

        if ((IsTileSolid_Virtual(playerObject->x + touchPointRightOffset, endPosition) != 0) || 
            (IsTileSolid_Virtual(playerObject->x + touchPointLeftOffset, endPosition) != 0))
        {
            startPosition >>= 5;
            startPosition <<= 5;

            playerObject->y = startPosition;
            playerObject->speedy = 0;

            if (onPhysicsCollision != NULL)
            {
                onPhysicsCollision();
            }
        }
    }

    playerObject->y += playerObject->speedy;

    JumpFallLeftRightCollisionCheck();
}

//#define PRINT_DEBUG_FALL_INFO 

void UpdateFallPhysics()
{
    UpdateObjectFall(playerObject, touchPointLeftOffset, touchPointRightOffset, PLAYER_PIXEL_HEIGHT, onPhysicsCollision);

    JumpFallLeftRightCollisionCheck();
}

void UpdateAutomated()
{
}

void PlayerManager_ReturnControlStartIdle()
{
    StartIdle();
}

void PlayerManager_ReturnControlStartFall()
{
    StartFall();
    playerObject->speedy = JUMP_SPEED;
}

void UpdateAutomatedPhysics()
{
    playerObject->x += playerObject->speedx;
    playerObject->y += playerObject->speedy;

    if (playerObject->speedy > JUMP_SPEED)
    {
        playerObject->speedy = JUMP_SPEED;
    }
}

void PlayerStartAutomatedWalk(s16 speed)
{
    playerData.playerState = PLAYERSTATE_AUTOMATED;

    StartWalkAnimation();

    playerObject->speedx = speed;

    // updates
    updateState = UpdateAutomated;
    updatePhysics = UpdateAutomatedPhysics;
}

void PlayerManager_StartAutomatedWalkLeft()
{
    PlayerStartAutomatedWalk(-MOVEMENT_SPEED);
}

void PlayerManager_StartAutomatedWalkRight()
{
    PlayerStartAutomatedWalk(MOVEMENT_SPEED);
}

void PlayerManager_StartAutomatedFall()
{
    playerData.playerState = PLAYERSTATE_AUTOMATED;

    StartJumpAnimation();

    playerObject->speedy = JUMP_SPEED;

    // updates
    updateState = UpdateAutomated;
    updatePhysics = UpdateAutomatedPhysics;
}

void UpdateDebugMovement()
{
    playerData.oldPositionX = playerObject->x;
    playerData.oldPositionY = playerObject->y;

    if (joystate & BUTTON_LEFT)
    {
        playerObject->speedx = -debugSettings.debugmovementspeed;
    }
    else if (joystate & BUTTON_RIGHT)
    {
        playerObject->speedx = debugSettings.debugmovementspeed;
    }
    else
    {
        playerObject->speedx = 0;
    }

    if (joystate & BUTTON_UP)
    {
        playerObject->speedy = -debugSettings.debugmovementspeed;
    }
    else if (joystate & BUTTON_DOWN)
    {
        playerObject->speedy = debugSettings.debugmovementspeed;
    }
    else
    {
        playerObject->speedy = 0;
    }

    playerObject->x += playerObject->speedx;
    playerObject->y += playerObject->speedy;
}

void UpdateOnGround()
{
    s16 yPosition = playerObject->y + PIXEL_TO_VIRTUAL(PLAYER_PIXEL_HEIGHT);

    playerData.onGround = (IsTileSolidOrSemiSolid_Virtual(playerObject->x + touchPointRightOffset, yPosition) != 0) || 
                          (IsTileSolid_Virtual(playerObject->x + touchPointLeftOffset, yPosition) != 0);
}

void UpdateHoleDetection()
{
}

//#define SHOW_STATE_CHANGES

#ifdef SHOW_STATE_CHANGES
static u16 oldState = -1;
#endif 

void PlayerUpdate(ObjectStruct* playerObject)
{
#ifdef SHOW_STATE_CHANGES
    if (oldState != playerData.playerState)
    {
        switch (playerData.playerState)
        {
        case PLAYERSTATE_IDLE:
            Print("PLAYERSTATE_IDLE");
            break;
        case PLAYERSTATE_JUMP:
            Print("PLAYERSTATE_JUMP");
            break;
        case PLAYERSTATE_WALK:
            Print("PLAYERSTATE_WALK");
            break;
        case PLAYERSTATE_FALL:
            Print("PLAYERSTATE_FALL");
            break;
        case PLAYERSTATE_AUTOMATED:
            Print("PLAYERSTATE_AUTOMATED");
            break;
        }

        oldState = playerData.playerState;
    }
#endif
    u16 disabled = playerData.disablePlayer;

    if (disabled == 1 &&
        playerData.playerState == PLAYERSTATE_AUTOMATED)
    {
        disabled = 0;
    }

    if (disabled == 0)
    {
        if (debugSettings.debugmovement == 1)
        {
            UpdateDebugMovement();
        }
        else
        {
            if (playerData.freezeTimer == 0)
            {
                playerData.oldPositionX = playerObject->x;
                playerData.oldPositionY = playerObject->y;

                updateState();
                updatePhysics();

                UpdateOnGround();

                if (playerData.playerState != PLAYERSTATE_AUTOMATED)
                {
                    UpdateHoleDetection();
                }

                if (playerData.invincibilityTimer > 0)
                {
                    playerData.invincibilityTimer--;
                }
            }
            else
            {
                playerData.freezeTimer--;

                if (playerData.freezeTimer == 0)
                {
                    if (playerData.onfreezeTimerDone != NULL)
                    {
                        playerData.onfreezeTimerDone();
                        playerData.onfreezeTimerDone = NULL;
                    }
                }

                // if we're frozen yet we're on a moving platform,
                // force update our position. Right now, this is 
                // for when the player down attacks and lands on a 
                // platform. 
                if (playerData.platform != NULL)
                {
                    playerObject->speedx = playerData.platform->speedx;
                    playerObject->speedy = playerData.platform->speedy;

                    playerObject->x += playerObject->speedx;
                    playerObject->y += playerObject->speedy;
                }
            }
        }
    }
}

void PlayerDestroy(ObjectStruct* playerObject)
{
}

void PlayerCollide(ObjectStruct* playerObject, ObjectStruct* collidedObject)
{
    if (debugSettings.debugmovement == 1)
        return;
}

const GGAnimation* oldAnimation = NULL;
u16 oldAnimationFrame = -1;

void PlayerManager_UpdateVDP()
{
    if ((oldAnimation != body.currentAnimation) ||
        (oldAnimationFrame != body.animationFrame))
    {
        VDPManager_CopyGGAnimationFrameToVDP(body.currentAnimation, body.animationFrame, &playerVDPInfo);
        oldAnimation = body.currentAnimation;
        oldAnimationFrame = body.animationFrame;
    }
}

void UpdateBodyAnimation()
{
    if (playerData.freezeTimer == 0)
    {
        // update Girl animation frame.
        body.animationSpeedCounter++;
        if (body.animationSpeedCounter > body.currentAnimation->frameDelayTime)
        {
            body.animationFrame++;
            body.animationSpeedCounter = 0;
        }

        if (body.animationFrame > body.currentAnimation->numberOfFrames - 1)
        {
            body.animationFrame = 0;
        }
    }

    GGAnimation_DrawPlayer(playerObject->spritedef.posx - 12, 
                           playerObject->spritedef.posy, 
                           body.currentAnimation, 
                           &playerVDPInfo, 
                           body.animationFrame,
                           playerObject->direction,
                           PAL0);
}

void PlayerDraw(ObjectStruct* playerObject)
{
    // update screen position
    playerObject->spritedef.posx = (s16)VIRTUAL_TO_PIXEL(playerObject->x - foregroundScrollData.scrollx_hires_clamped);
    playerObject->spritedef.posy = (s16)VIRTUAL_TO_PIXEL(playerObject->y - foregroundScrollData.scrolly_hires_clamped);


    // debugging
    if (debugSettings.showCollisionBox == 1)
    {
        MyVDP_setSprite(playerObject->spritedef.posx - 4, playerObject->spritedef.posy + 24, SPRITE_SIZE(1,1), 1);
        MyVDP_setSprite(playerObject->spritedef.posx - 4, playerObject->spritedef.posy + 16, SPRITE_SIZE(1,1), 1);
        MyVDP_setSprite(playerObject->spritedef.posx - 4, playerObject->spritedef.posy + 8, SPRITE_SIZE(1,1), 1);
        MyVDP_setSprite(playerObject->spritedef.posx - 4, playerObject->spritedef.posy, SPRITE_SIZE(1,1), 1);
    }

    if (playerData.invincibilityTimer & 1)
    {
        return;
    }

    UpdateBodyAnimation();
}

void PlayerManager_InitPosition(s16 x, s16 y)
{
    playerData.playerLevelStartX = x;
    playerData.playerLevelStartY = y;
    playerData.oldPositionX = x;
    playerData.oldPositionY = y;
}

void PlayerManager_SetupForGame()
{
    playerObject = &player;
    playerObject->updateFunction    = PlayerUpdate;
    playerObject->destroyFunction   = PlayerDestroy;
    playerObject->drawFunction      = PlayerDraw;
    playerObject->collisionFunction = PlayerCollide;
    playerObject->objectState       = OBJECTSTATE_ACTIVE;

    playerObject->objectType        = TYPE_PLAYER;
    playerObject->spritedef.size    = SPRITE_SIZE(3,4);
    playerObject->damage            = 1;
    playerObject->healthPoints      = MAX_HEALTH_POINTS;

    playerObject->rectLeft          = PIXEL_TO_VIRTUAL(-4);
    playerObject->rectRight         = PIXEL_TO_VIRTUAL(3);
    playerObject->rectTop           = PIXEL_TO_VIRTUAL(0);
    playerObject->rectBottom        = PIXEL_TO_VIRTUAL(PLAYER_PIXEL_HEIGHT - 1);
     
    playerData.playerLevelStartX    = 0;
    playerData.playerLevelStartY    = 0;
    playerData.oldPositionX         = 0;
    playerData.oldPositionY         = 0;
}

void PlayerManager_AfterDeath()
{
    playerObject->healthPoints      = MAX_HEALTH_POINTS;
}

void PlayerManager_SetupLife()
{
    playerObject->x                 = playerData.playerLevelStartX;
    playerObject->y                 = playerData.playerLevelStartY;
    playerObject->speedx            = 0;
    playerObject->speedy            = 0;
    playerObject->direction         = DIRECTION_RIGHT;


    playerData.invincibilityTimer   = 0;
    playerData.freezeTimer          = 0;
    playerData.onfreezeTimerDone    = NULL;
    playerData.playerState          = 0;
    playerData.disablePlayer        = 0;
    playerData.platform             = NULL;
    playerData.oldPositionX         = playerData.playerLevelStartX;
    playerData.oldPositionY         = playerData.playerLevelStartY;
    playerData.order                = 1;

    deathWaitTime = 0;
    jumpCounter = 0;

    oldAnimation = NULL;
    oldAnimationFrame = -1;

    StartIdle();
}
