#include "ObjectMovement.h"

#include "..\Collisions.h"
#include "..\LevelManager.h"
#include "..\MathUtils.h"
#include "..\TileHeightMaps.h"

void UpdateObjectLeftRightWalkPhysics(ObjectStruct* objectStruct, 
                                      s16 (*leftRightCollisionCheck)(ObjectStruct* objectStruct, s16 speedx), 
                                      s16 sensorOffsetInPixels)
{
    s16 sensorOffsetInVirtual = PIXEL_TO_VIRTUAL(sensorOffsetInPixels);
    s16 sensorYMinusOne = objectStruct->y + PIXEL_TO_VIRTUAL(sensorOffsetInPixels - 1);
    u16 currenttiletype = GetTileType_Virtual(objectStruct->x, sensorYMinusOne);
    u16 currentTile = GetTile_Virtual(objectStruct->x, sensorYMinusOne);
    u16 isCurrentTileFlipped = IS_FLIPPED_HORIZONTALLY(currentTile);

    // determine the Walking speed from the slope we're on
    s16 effectiveXSpeed = objectStruct->speedx;

    if (currenttiletype & TILE_SLOPE)
    {
        u16 slopeHeightMapIndex = ((currenttiletype & 0xf0) >> 4) - 1;

        // going left
        if (objectStruct->speedx < 0)
        {
            if (isCurrentTileFlipped)
            {
                effectiveXSpeed -= slopeClimbUpSpeedModifier[slopeHeightMapIndex];
            }
            else
            {
                effectiveXSpeed += slopeClimbUpSpeedModifier[slopeHeightMapIndex];
            }
        }
        // going right
        else if (objectStruct->speedx > 0)
        {
            if (isCurrentTileFlipped)
            {
                effectiveXSpeed -= slopeClimbUpSpeedModifier[slopeHeightMapIndex];
            }
            else
            {
                effectiveXSpeed += slopeClimbUpSpeedModifier[slopeHeightMapIndex];
            }
        }
    }

    effectiveXSpeed = leftRightCollisionCheck(objectStruct, effectiveXSpeed);

    if (effectiveXSpeed == 0)
    {
        return;
    }

    // insert x position processing here.
    objectStruct->x += effectiveXSpeed;

    // slope processing
    u16 pixelx = VIRTUAL_TO_PIXEL(objectStruct->x);
    u16 sensortilex = VIRTUAL_TO_TILE(objectStruct->x);
    u16 sensortiley = VIRTUAL_TO_TILE(sensorYMinusOne);

    u16 nexttiletype = GetTileType_Virtual(objectStruct->x, sensorYMinusOne);

    s16 heightmapindex = 0;


    if (objectStruct->speedx != 0)
    {
        // if we're on a slope
        //      if we're moving up (ie going right on an up/right slope) (up/right == down/left)
        //          - check the tile over in the next tile  [1][2]
        //          - if solid                     [2]
        //              check one tile up       [1][X]
        //              if tile is empty
        //                  handle reaching the top
        //                  return
        //          - if tile is slope
        //              handle going to next slope
        //              return

        if (currenttiletype & TILE_SLOPE)
        {
            if ((nexttiletype & TILE_SOLID))
            {
                sensortiley--;
                nexttiletype = GetTileType_Tile(sensortilex, sensortiley);
            }

            // we're on another slope
            if (nexttiletype & TILE_SLOPE)
            {
                // get the appropriate tile height map
                u16 slopeHeightMapIndex = ((nexttiletype & 0xf0) >> 4) - 1;
                const u16* tileHeightMap = slopeHeightMaps[slopeHeightMapIndex];

                // determine the index to use in the tile height map
                u16 nextTile = GetTile_Tile(sensortilex, sensortiley);
                u16 isNextTileFlipped = IS_FLIPPED_HORIZONTALLY(nextTile);

                if (!isNextTileFlipped)
                {
                    heightmapindex = pixelx & 7;
                }
                else
                {
                    heightmapindex = 7 - (pixelx & 7);
                }

                objectStruct->y = TILE_TO_VIRTUAL(sensortiley + 1) - sensorOffsetInVirtual - PIXEL_TO_VIRTUAL(tileHeightMap[heightmapindex]);
            }
            // we're reached the top
            else if (nexttiletype == TILE_NONE)
            {
                objectStruct->y = TILE_TO_VIRTUAL(sensortiley + 1) - sensorOffsetInVirtual;
            }
        }
        //  - check if we're going down
        else if (nexttiletype == TILE_NONE)
        {
            sensortiley++;
            nexttiletype = GetTileType_Tile(sensortilex, sensortiley);

            if (nexttiletype & TILE_SLOPE)
            {
                // get the appropriate tile height map
                u16 slopeHeightMapIndex = ((nexttiletype & 0xf0) >> 4) - 1;
                const u16* tileHeightMap = slopeHeightMaps[slopeHeightMapIndex];

                // determine the index to use in the tile height map
                u16 nextTile = GetTile_Tile(sensortilex, sensortiley);
                u16 isNextTileFlipped = IS_FLIPPED_HORIZONTALLY(nextTile);

                if (!isNextTileFlipped)
                {
                    heightmapindex = pixelx & 7;
                }
                else
                {
                    heightmapindex = 7 - (pixelx & 7);
                }

                objectStruct->y = TILE_TO_VIRTUAL(sensortiley + 1) - sensorOffsetInVirtual - PIXEL_TO_VIRTUAL(tileHeightMap[heightmapindex]);
            }
        }
    }
}

void UpdateObjectFall(ObjectStruct* objectStruct, 
                      s16 leftSensorX, 
                      s16 rightSensorX, 
                      s16 yOffset,
                      void (*onGroundTouchCallback)(ObjectStruct* objectStruct))
{
    // sensor locations on x axis
    s16 leftSensor = objectStruct->x + leftSensorX;
    s16 rightSensor = objectStruct->x + rightSensorX;

    // if the bottom pixel is touching a semi-solid tile, then
    // we don't care about collisions as we've passed through it
    s16 startPosition = objectStruct->y + PIXEL_TO_VIRTUAL(yOffset - 1);

    if ((IsTileSemiSolid_Virtual(rightSensor, startPosition) != 0) || 
        (IsTileSemiSolid_Virtual(leftSensor, startPosition) != 0))
    {
        goto do_physics;
    }

    // get the pixel immediately underneath the player
    startPosition = objectStruct->y + PIXEL_TO_VIRTUAL(yOffset);

    // only care if we're on the map
    if (startPosition > TILE_TO_VIRTUAL(currentForegroundMap->height))
    {
        goto do_physics;
    }

    s16 endPosition = startPosition + objectStruct->speedy;

    // only care if the end position is on the map;
    if (endPosition > TILE_TO_VIRTUAL(currentForegroundMap->height) ||
        endPosition < 0)
    {
        goto do_physics;
    }

    if ((IsTileSolidOrSemiSolid_Virtual(rightSensor, endPosition) != 0) || 
        (IsTileSolidOrSemiSolid_Virtual(leftSensor, endPosition) != 0))
    {
        // touched a solid tile
        endPosition >>= 5;
        endPosition <<= 5;
        objectStruct->y = endPosition - PIXEL_TO_VIRTUAL(yOffset);
        objectStruct->speedy = 0;

        if (onGroundTouchCallback != NULL)
        {
            onGroundTouchCallback(objectStruct);
        }
    }
    else // check if we've collided with a slope.
    {
        u16 rawtile = GetTile_Virtual(objectStruct->x, endPosition);
        u16 tile = rawtile & TILEINDEX_BITMASK;
        u16 tileType = ARRAY_ITEM_VALUE(currentForegroundMap->tilesetAttributes, tile);

        if (tileType & TILE_SLOPE) // check if we've touched a slope.
        {
                        
            // get the appropriate tile height map
            s16 pixelx = VIRTUAL_TO_PIXEL(objectStruct->x);

            u16 slopeHeightMapIndex = ((tileType & 0xf0) >> 4) - 1;
            const u16* tileHeightMap = slopeHeightMaps[slopeHeightMapIndex];

            // determine the height index to use in the tile height map
            s16 heightmapheightindex = 0;
                    
            if (IS_FLIPPED_HORIZONTALLY(rawtile))
            {
                heightmapheightindex = 7 - (pixelx & 7);
            }
            else
            {
                heightmapheightindex = pixelx & 7;
            }

            s16 heightValue = tileHeightMap[heightmapheightindex];

            // see if we're touching the slope.
            s16 slopetiley = VIRTUAL_TO_PIXEL(endPosition) & 7;
            s16 heightvaluey = (7 - heightValue);

            if (slopetiley > heightvaluey)
            {
                // slope hit.
                objectStruct->y = endPosition - PIXEL_TO_VIRTUAL(yOffset);
                objectStruct->speedy = 0;
                if (onGroundTouchCallback != NULL)
                {
                    onGroundTouchCallback(objectStruct);
                }
            }
        }
    }

do_physics:
    objectStruct->y += objectStruct->speedy;

}

