#include "Pingouin.h"

#include "ObjectCommon.h"
#include "..\HUDManager.h"
#include "..\TileHeightMaps.h"
#include "..\StateManager.h"
#include "..\generated\AllLevels.h"

#define PINGOUIN_IDLE           0
#define PINGOUIN_WALK           1
#define PINGOUIN_JUMP           2
#define PINGOUIN_FALL           3
#define PINGOUIN_ENTERINGHOUSE  4
#define PINGOUIN_ENDING         5

#define PINGOUIN_IDLE_ANIMATION     0
#define PINGOUIN_WALK_ANIMATION     1
#define PINGOUIN_JUMP_ANIMATION     2
#define PINGOUIN_FALL_ANIMATION     2
#define PINGOUIN_HAPPY_ANIMATION    3

#define PINGOUIN_WALKSPEED 4

void PingouinStartEnteringHouse(ObjectStruct* object);

ObjectStruct* PingouinCreate(s16 x, s16 y)
{
    ObjectStruct* object = ObjectManager_CreateObject(TYPE_PICKUP,
                                                      PingouinUpdate, 
                                                      PingouinDestroy, 
                                                      PingouinDraw,
                                                      PingouinCollide);

    if (!object)
    {
#ifdef DEBUG_OBJECTS
        KDebug_Alert("Pingouin could not be created!");
#endif
        return NULL;
    }

    object->x                           = x;
    object->y                           = y;
    object->animation                   = NULL;
    object->speedx                      = 0;
    object->speedy                      = 0;
    object->rectLeft                    = PIXEL_TO_VIRTUAL(-4);
    object->rectRight                   = PIXEL_TO_VIRTUAL(4);
    object->rectTop                     = 0;
    object->rectBottom                  = PIXEL_TO_VIRTUAL(16);
    object->healthPoints                = 1;
    object->damage                      = 0;
    object->animationFrame              = 0;
    object->animationSubFrameCounter    = 0;
    object->direction                   = 0;
    object->palette                     = PAL1;
    object->aiState                     = PINGOUIN_IDLE;

    ComputeStandardActiveArea(object);

#ifdef DEBUG_OBJECTS
    KDebug_Alert("PingouinCreate");
#endif

    return object;
}

void PingouinStartWalk(ObjectStruct* object)
{
    //Print("StartWalk");
    object->aiState = PINGOUIN_WALK;
    object->animationFrame = 0;

    const GGAnimation* const* animations = (const GGAnimation* const*)object->extra2;
    object->animation = animations[PINGOUIN_WALK_ANIMATION];
}

void PingouinStartJump(ObjectStruct* object)
{
    //Print("StartJump");
    object->aiState = PINGOUIN_JUMP;
    object->speedy = -20;
    object->animationFrame = 0;

    const GGAnimation* const* animations = (const GGAnimation* const*)object->extra2;
    object->animation = animations[PINGOUIN_JUMP_ANIMATION];
}

void PingouinStartFall(ObjectStruct* object)
{
    //Print("StartFall");
    object->aiState = PINGOUIN_FALL;
    object->speedy = 0;
    object->animationFrame = 0;

    const GGAnimation* const* animations = (const GGAnimation* const*)object->extra2;
    object->animation = animations[PINGOUIN_JUMP_ANIMATION];
}

void PingouinStartEnding(ObjectStruct* object)
{
    object->aiState = PINGOUIN_ENDING;
    object->lifetime = 0;
    object->animationFrame = 0;

    const GGAnimation* const* animations = (const GGAnimation* const*)object->extra2;
    object->animation = animations[PINGOUIN_HAPPY_ANIMATION];
}

void PingouinStartEnteringHouse(ObjectStruct* object)
{
    object->aiState = PINGOUIN_ENTERINGHOUSE;
    object->lifetime = 0;
    object->animationFrame = 0;

    const GGAnimation* const* animations = (const GGAnimation* const*)object->extra2;
    object->animation = animations[PINGOUIN_HAPPY_ANIMATION];

    PlayerManager_Happy();
}

void PingouinJumpFallLeftRightCollisionCheck(ObjectStruct* object)
{
    s16 endXPosition = object->x + object->speedx;
    s16 objectY = object->y + PIXEL_TO_VIRTUAL(object->animation->height - 1);

    if (object->speedx > 0)
    {
        endXPosition += PIXEL_TO_VIRTUAL(4);

        if (IsTileSolid_Virtual(endXPosition, objectY) != 0)
        {
            endXPosition >>= 5;
            endXPosition <<= 5;

            object->x = endXPosition - PIXEL_TO_VIRTUAL(4);
            object->speedx = 0;
        }
    }
    else if (object->speedx < 0)
    {
        endXPosition -= PIXEL_TO_VIRTUAL(4);

        if (IsTileSolid_Virtual(endXPosition, objectY) != 0)
        {
            endXPosition >>= 5;
            endXPosition <<= 5;
            endXPosition += 32;

            object->x = endXPosition + PIXEL_TO_VIRTUAL(4);
            object->speedx = 0;
        }
    }
}

s16 PingouinWalkLeftRightCollisionCheck(ObjectStruct* object, s16 speedx)
{
    s16 endXPosition = object->x + speedx;
    s16 objectY = object->y + PIXEL_TO_VIRTUAL(object->animation->height - 9);

    if (speedx > 0)
    {
        endXPosition += PIXEL_TO_VIRTUAL(4);

        if (IsTileSolid_Virtual(endXPosition, objectY) != 0)
        {
            endXPosition >>= 5;
            endXPosition <<= 5;

            object->x = endXPosition - PIXEL_TO_VIRTUAL(4);
            speedx = 0;

            PingouinStartJump(object);
        }
    }
    else if (speedx < 0)
    {
        endXPosition -= PIXEL_TO_VIRTUAL(4);

        if (IsTileSolid_Virtual(endXPosition, objectY) != 0)
        {
            endXPosition >>= 5;
            endXPosition <<= 5;
            endXPosition += 32;

            object->x = endXPosition + PIXEL_TO_VIRTUAL(4);
            speedx = 0;

            PingouinStartJump(object);
        }
    }

    return speedx;
}

static void UpdateGroundCheck(ObjectStruct* object)
{
    // ground check
    s16 ySensorPosition = object->y + PIXEL_TO_VIRTUAL(object->animation->height);

    if (ySensorPosition < 0)
    {
        ySensorPosition = 0;
    }

    if ((GetTileType_Virtual(object->x + object->rectRight, ySensorPosition) == TILE_NONE) &&
        (GetTileType_Virtual(object->x + object->rectLeft, ySensorPosition) == TILE_NONE))
    {
        PingouinStartFall(object);
    }
}

void PingouinUpdateFallPhysics(ObjectStruct* object)
{
    object->speedy++;

    if (object->speedy > 8)
    {
        object->speedy = 8;
    }

    UpdateObjectFall(object, 0, 0, object->animation->height, PingouinStartWalk);

    PingouinJumpFallLeftRightCollisionCheck(object);
}

void PingouinUpdateJump(ObjectStruct* object)
{
    PingouinJumpFallLeftRightCollisionCheck(object);

    object->speedy++;
    if (object->speedy == 0)
    {
        PingouinStartFall(object);
    }
}

void UpdateEnteringHouse(ObjectStruct* object)
{
    object->x += object->speedx;

    if (VIRTUAL_TO_PIXEL(object->x) < 32)
    {
        object->x = PIXEL_TO_VIRTUAL(32);
        object->speedx = 0;    
    }

    if (object->lifetime > 250)
    {
        const Level* nextLevel = &simple;

        if (currentLevel->nextlevel != NULL)
        {
            nextLevel = currentLevel->nextlevel;
        }

        LevelManager_SetLevel(nextLevel);
        StateManager_ChangeToState(LevelManager_StartLevel);
    }
}


void PingouinUpdatePhysics(ObjectStruct* object)
{
    s16 targetX = 0;

    if (playerObject->direction == 0)
    {
        targetX = playerObject->x - PIXEL_TO_VIRTUAL(16) * object->extra1;
    }
    else
    {
        targetX = playerObject->x + PIXEL_TO_VIRTUAL(16) * object->extra1;
    }

    s16 difference = mathutils_abs(object->x - targetX);

    if (difference > PINGOUIN_WALKSPEED)
    {
        if (targetX > object->x)
        {
            object->speedx = PINGOUIN_WALKSPEED;
        }
        else
        {
            object->speedx = -PINGOUIN_WALKSPEED;
        }
    }
    else
    {
        object->speedx = 0;
    }

    if (object->aiState == PINGOUIN_WALK)
    {
        UpdateObjectLeftRightWalkPhysics(object, PingouinWalkLeftRightCollisionCheck, 24);

        UpdateGroundCheck(object);
    }
    else if (object->aiState == PINGOUIN_FALL)
    {
        PingouinUpdateFallPhysics(object);

        object->x += object->speedx;
    }
    else if (object->aiState == PINGOUIN_JUMP)
    {
        PingouinUpdateJump(object);

        object->x += object->speedx;
    }

    
    object->y += object->speedy;
}

void PingouinUpdate(ObjectStruct* object)
{
    if (isEndingScreen)
    {
        if (object->aiState != PINGOUIN_ENDING)
        {
            PingouinStartEnding(object);
        }
        else
        {

        }
    }
    else
    {
        if (object->aiState != PINGOUIN_ENTERINGHOUSE)
        {
            if (object->aiState != PINGOUIN_IDLE)
            {
                PingouinUpdatePhysics(object);
            }

            if (object->x < TILE_TO_VIRTUAL(14))
            {
                PingouinStartEnteringHouse(object);
            }
        }
        else
        {
            UpdateEnteringHouse(object);
        }
    
    }


    if (object->speedx != 0)
    {
        object->direction = (object->speedx < 0);
    }

    object->spritedef.posx = VIRTUAL_TO_PIXEL(object->x) - (object->animation->width >> 1) - foregroundScrollData.scrollx_vdp;
    object->spritedef.posy = VIRTUAL_TO_PIXEL(object->y) - foregroundScrollData.scrolly_vdp;
}

void PingouinDestroy(ObjectStruct* object)
{
#ifdef DEBUG_OBJECTS
    KDebug_Alert("PingouinDestroy");
#endif
}

void PingouinDraw(ObjectStruct* object)
{
    object->animationSubFrameCounter++;

    if (object->animationSubFrameCounter > object->animation->frameDelayTime)
    {
        object->animationFrame++;
        object->animationSubFrameCounter = 0;
    }

    if (object->animationFrame > object->animation->numberOfFrames - 1)
    {
        object->animationFrame = 0;
    }

    GGAnimation_DrawObject(object);
}

void PingouinCollide(ObjectStruct* object, ObjectStruct* collidedObject)
{
    if (object->aiState == PINGOUIN_IDLE)
    {
        PingouinStartJump(object);

        object->extra1 = playerData.order;
        playerData.order++;
    }
}
